import warnings
from collections.abc import Callable, Iterable
from typing import Any, Never


class Maybe[T]:
    """
    Wraps a value that may be ``T`` or ``None``, providing methods for conditionally using that value or
    short-circuiting to ``None`` without longer checks.
    """
    __match_args__ = ('val',)

    def __init__(self, val: T | None) -> None:
        warnings.warn(
            'Direct instancing of Maybe() is not recommended as of v0.5.0, use the maybe() function instead',
            stacklevel=2,
        )
        self.val = val

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.val!r})'

    def __bool__(self) -> bool:
        return self.val is not None

    def __eq__(self, other: object) -> bool:
        """
        Returns ``False`` if the compared object is not a ``Maybe`` instance, otherwise compares their wrapped values.
        """
        if not isinstance(other, Maybe):
            return False
        return self.val == other.val

    def __hash__(self) -> int:
        return self.val.__hash__()

    @staticmethod
    def cat(vals: 'Iterable[Maybe[T]]') -> list[T]:
        """
        Takes an iterable of ``Maybe`` and returns a list of the unwrapped values of all ``Some`` objects.

        >>> vals = [maybe(5), maybe(None), maybe(10), maybe(None)]
        >>> assert vals == [Some(5), Nothing, Some(10), Nothing]
        >>> assert Maybe.cat(vals) == [5, 10]
        """
        return [i.unwrap() for i in vals if i]

    @staticmethod
    def int(val: Any) -> 'Maybe[int]':  # noqa: ANN401
        """
        Attempts to convert ``val`` to an ``int``, returning a ``Some``-wrapped ``int`` if successful, or
        ``Nothing`` on failure.
        """
        try:
            i: int = int(val)
            return Some(i)
        except ValueError:
            return Nothing

    @staticmethod
    def map[A, B](fn: 'Callable[[A], Maybe[B]]', vals: Iterable[A]) -> list[B]:
        """Maps ``fn`` onto ``vals``, taking the unwrapped values of ``Some``s and discarding ``Nothing``s."""
        return [i.unwrap() for i in map(fn, vals) if i]

    def attr[V](self, name: str, typ: type[V] | None = None, *, err: bool = False) -> 'Maybe[V]':
        """
        Attempts to access an attribute ``name`` on the wrapped object, returning a ``Some`` instance wrapping the
        the value if it exists, or ``Nothing`` otherwise.

        :param typ: Specifies the generic type of the resulting ``Maybe``. Note that the potential value returned by
            this method will not be coerced to the given type at runtime; this argument is only for typing purposes.
        :param err: If ``True``, ``AttributeError`` is raised instead of returning ``Nothing`` if ``name`` does not
            exist.
        """
        return Some(getattr(self.val, name)) if err else maybe(getattr(self.val, name, None))

    def attr_or[V](self, name: str, default: V) -> V:
        """
        Similar to the ``attr`` method, but unwraps the result if the attribute exists or returns the required default
        value otherwise.
        """
        try:
            return self.attr(name, err=True).unwrap()
        except AttributeError:
            return default

    def get[V](self,
            accessor: Any,  # noqa: ANN401
            _typ: type[V] | None = None,
            *,
            err: bool = False,
            default: V | None = None,
        ) -> 'Maybe[V]':
        """
        Attempts to access an item by ``accessor`` on the wrapped object, assuming the wrapped value implements
        ``__getitem__``. If it does not, or if the value does not exist (list index out of range, key does not exist on
        a dictionary, etc.), ``Nothing`` is returned.

        :param typ: Specifies the generic type of the resulting ``Maybe``. Note that the potential value returned by
            this method will not be coerced to the given type at runtime; this argument is only for typing purposes.
        :param err: By default, ``IndexError`` and ``KeyError`` are not raised when ``__getitem__`` is called on the
            wrapped value, and ``Nothing`` is returned instead. Setting ``err`` to ``True`` allows these errors to
            be raised as they normally would. Note that if ``__getitem__`` did not exist on the wrapped value in the
            first placed (such as with ``Nothing``), no error is raised, and ``Nothing`` is returned regardless.
        :param default: Specifies an alternate value to return a ``Some`` of instead of returning ``Nothing``.
        """
        if hasattr(self.val, '__getitem__'):
            try:
                return Some(self.val.__getitem__(accessor)) # type: ignore
            except (IndexError, KeyError):
                if err:
                    raise
        return maybe(default)

    def then[R](self, func: Callable[[T], R]) -> R | None:
        """
        Calls ``func`` with the wrapped value as the argument and returns its value, or returns ``None`` if the wrapped
        value is ``None``.

        :param func: A ``Callable`` which takes a type of the possible wrapped value (``T``) and can return any type
            (``R``).
        """
        return func(self.val) if self.val is not None else None

    def this_or(self, other: T) -> 'Maybe[T]':
        """Returns the original wrapped value if not ``None``, otherwise returns a ``Some``-wrapped ``other``."""
        return self if self else Some(other)

    def unwrap(self,
            exc: Exception | Callable[..., Never] | None = None,
            *exc_args: object,
        ) -> T:
        """
        Returns the wrapped value if it is not ``None``, otherwise raises ``ValueError`` by default.

        :param exc: The exception to raise if the wrapped value is ``None``. Can be either an ``Exception`` object, or a
            ``Callable`` which takes any arguments and does not return. If given ``None``, the default behavior is to
            raise a ``ValueError`` with the message ``Maybe[<type>] unwrapped into None``.
        :param exc_args: Arguments to call ``exc`` with, if ``exc`` is a ``Callable``. Otherwise, this argument is not
            used.
        """
        if self.val is None:
            if isinstance(exc, Exception):
                raise exc
            if isinstance(exc, Callable):
                exc(*exc_args)
            raise ValueError(f'Maybe[{T.__name__}] unwrapped into None')
        return self.val

    def unwrap_or(self, other: T) -> T:
        """Returns the wrapped value if it is not ``None``, otherwise returns ``other``."""
        return self.val if self.val is not None else other

class Some[T](Maybe[T]):
    def __init__(self, val: T) -> None:
        self.val = val

    def __bool__(self) -> bool:
        return True

class _Nothing[T](Maybe[T]):
    __match_args__ = ()

    def __init__(self, _: None = None) -> None:
        self.val = None

    def __repr__(self) -> str:
        return 'Nothing'

    def __bool__(self) -> bool:
        return False

Nothing = _Nothing()

def maybe[T](val: T | None, predicate: Callable[[T], bool] = lambda v: v is not None) -> Maybe[T]:
    """
    Returns a ``Some`` instance wrapping ``val`` if either ``val`` is not ``None`` or ``predicate(val)`` is ``True``,
    otherwise returns the ``Nothing`` singleton.

    :param val: A value to wrap.
    :param predicate: An optional function that takes ``val`` and, if it returns ``False``, discards ``val``
        and returns a ``Nothing`` instance. Regardless of the predicate, ``Nothing`` is always returned if
        ``val`` is ``None``.
    """
    return Nothing if (val is None) or not predicate(val) else Some(val)



import re
from collections.abc import Callable
from copy import deepcopy
from dataclasses import dataclass
from string import ascii_lowercase
from types import EllipsisType
from typing import Any

import pytest

from maybetype import Maybe, Nothing, Some, maybe

ALPHANUMERIC: str = ascii_lowercase + '0123456789'
MAYBE_UNWRAP_NONE_REGEX: re.Pattern[str] = re.compile(r"Maybe\[.*\] unwrapped into None")

def test_maybe_none_unwrap_error() -> None:
    m_none: Maybe[Any] = Nothing
    assert bool(m_none) is False
    with pytest.raises(ValueError, match=MAYBE_UNWRAP_NONE_REGEX):
        m_none.unwrap()
    with pytest.raises(TypeError, match='Custom error message'):
        m_none.unwrap(exc=TypeError('Custom error message'))

def test_maybe_none_is_nothing() -> None:
    assert maybe(None) is Nothing

def test_maybe_this_or() -> None:
    assert Maybe.int('10').this_or(0).unwrap() == 10  # noqa: PLR2004
    assert Maybe.int('ten').this_or(0).unwrap() == 0
    with pytest.raises(ValueError, match=MAYBE_UNWRAP_NONE_REGEX):
        Maybe.int('ten').unwrap()

@pytest.mark.parametrize(('val', 'default'),
    [
        (5, 10),
        (5, None),
        ('string', 'fallback'),
        ('string', None),
    ],
)
def test_maybe_unwrap_or(val: object, default: object) -> None:
    assert maybe(val).unwrap_or(default) == val
    assert maybe(None).unwrap_or(default) == default

@pytest.mark.parametrize(('val', 'then_fn'),
    [
      # (wrapped value, (.then() function, expected return val), expected wrapped value after function)
        (0,                (lambda i: i * 10, 0, ...)),
        (1,                (lambda i: i * 10, 10, ...)),
        ('',               (lambda s: s.upper(), '', ...)),
        ('string',         (lambda s: s.upper(), 'STRING', ...)),
        ('a,b,c',          (lambda s: s.split(','), ['a', 'b', 'c'], ...)),
        ([1, 2, 3],        (lambda l: l.append(4), None, [1, 2, 3, 4])), # noqa: E741
        ({'a': 1, 'b': 2}, (lambda d: d.get('a'), 1, ...)),
    ],
    ids=[
        'int_zero',
        'int_one',
        'str_empty',
        'str_nonempty',
        'str_split',
        'list',
        'dict',
    ],
)
def test_maybe_then[T, R, A](val: T, then_fn: tuple[Callable[[T], R], R, A | EllipsisType]) -> None:
    m: Maybe[T] = maybe(val)
    assert bool(m) is True
    assert m.unwrap() == val
    m_before = deepcopy(m.val)
    assert m.then(then_fn[0]) == then_fn[1]
    assert m.unwrap() == (then_fn[2] if then_fn[2] is not Ellipsis else m_before)
    assert maybe(None).then(then_fn[0]) is None

def test_maybe_attr() -> None:
    @dataclass
    class A:
        x: int

    @dataclass
    class B(A):
        y: float

    m_none: Maybe[A] = maybe(None)
    assert m_none.attr('x').val is None
    assert m_none.attr_or('x', 2) == 2  # noqa: PLR2004

    m_a: Maybe[A] = maybe(A(1))
    assert m_a.attr('x').unwrap() == 1
    assert m_a.attr_or('x', 2) == 1
    assert m_a.attr('y').val is None
    assert m_a.attr_or('y', 2) == 2  # noqa: PLR2004

    m_b: Maybe[B] = maybe(B(1, 2.0))
    assert m_b.attr('x').unwrap() == 1
    assert m_b.attr_or('x', 2) == 1
    assert m_b.attr('y').unwrap() == 2.0  # noqa: PLR2004
    assert m_b.attr_or('y', 3) == 2.0  # noqa: PLR2004

@pytest.mark.parametrize(('val', 'accessor', 'result'),
    [
        (None,             1,   maybe(None)),
        ([1, 2, 3],        1,   maybe(2)),
        ([1, 2, 3],        3,   maybe(None)),
        ([],               1,   maybe(None)),
        ({'a': 1, 'b': 2}, 'a', maybe(1)),
        ({'a': 1, 'b': 2}, 'c', maybe(None)),
        ({},               'a', maybe(None)),
    ],
    ids=[
        'none',
        'list_populated',
        'list_populated_out_of_range',
        'list_empty',
        'dict_populated',
        'dict_populated_no_key',
        'dict_empty',
    ],
)
def test_maybe_get(val: object, accessor: object, result: object) -> None:
    m: Maybe = maybe(val)
    assert m.get(accessor) == result

def test_maybe_cat() -> None:
    assert Maybe.cat(map(Maybe.int, ALPHANUMERIC)) == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

def test_maybe_cat_failure() -> None:
    with pytest.raises(AttributeError, match='has no attribute \'unwrap\''):
        Maybe.cat([1, 2, 3]) # type: ignore

def test_maybe_map() -> None:
    assert Maybe.map(Maybe.int, ALPHANUMERIC) == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

def is_valid_uuid(s: str) -> bool:
    return re.match(r"[0-9a-f]{8}(?:-[0-9a-f]{4}){3}-[0-9a-f]{12}|[0-9a-f]{32}", s) is not None

@pytest.mark.parametrize(('value', 'predicate', 'expected_bool'),
    [
        (0, lambda a: a > 0, False),
        ([], lambda a: len(a) > 0, False),
        ([], lambda a: 'x' in a, False),
        ({}, lambda a: len(a) > 0, False),
        ({}, lambda a: 'x' in a, False),
        ('3b1bcc3a-41d5-49a5-8273-10cc605e31f9', is_valid_uuid, True),
        ('3b1bcc3a41d549a5827310cc605e31f9', is_valid_uuid, True),
        ('qwertyuiopasdfghjklzxcvbnm', is_valid_uuid, False),
        ('nf0cmmdq-l0gt-rq5a-upry-706trht3ocv9', is_valid_uuid, False),
    ],
)
def test_maybe_with_predicate[T](value: T, predicate: Callable[[T], bool], expected_bool: bool) -> None:
    assert bool(maybe(value, predicate)) is expected_bool
    if expected_bool is False:
        assert maybe(value, predicate) is Nothing

@pytest.mark.parametrize(('value', 'predicate', 'expected'),
    [
        (0, lambda a: a > 0, None),
        ([], lambda a: len(a) > 0, None),
        ([], lambda a: 'x' in a, None),
        ({}, lambda a: len(a) > 0, None),
        ({}, lambda a: 'x' in a, None),
        ('3b1bcc3a-41d5-49a5-8273-10cc605e31f9', is_valid_uuid, ...),
        ('3b1bcc3a41d549a5827310cc605e31f9', is_valid_uuid, ...),
        ('qwertyuiopasdfghjklzxcvbnm', is_valid_uuid, None),
        ('nf0cmmdq-l0gt-rq5a-upry-706trht3ocv9', is_valid_uuid, None),
    ],
)
def test_maybe_pattern_matching[T](value: T, predicate: Callable[[T], bool], expected: T | None) -> None:
    if expected is Ellipsis:
        expected = value

    match maybe(value, predicate):
        case Some(value):
            result = value
        case _:
            result = None

    assert result == expected
